//! Code which enables sending messages between processes in the blocktree system.
use btlib::{crypto::rand_array, error::BoxInIoErr, Principal, Result};
use btserde::{read_from, write_to};
use bytes::{BufMut, BytesMut};
use core::{
    future::Future,
    pin::Pin,
    task::{Context, Poll},
};
use futures::{
    sink::{Send, Sink},
    stream::Stream,
    SinkExt, StreamExt,
};
use lazy_static::lazy_static;
use serde::{Deserialize, Serialize};
use std::{io, marker::PhantomData, net::Shutdown, path::PathBuf};
use tokio::{
    io::{AsyncRead, AsyncWrite, ReadBuf},
    net::UnixDatagram,
};
use tokio_util::codec::{Decoder, Encoder, FramedRead, FramedWrite};
use zerocopy::FromBytes;

pub use private::*;

mod private {

    use super::*;

    lazy_static! {
        /// The default directory in which to place blocktree sockets.
        static ref SOCK_DIR: PathBuf = {
            let mut path: PathBuf = std::env::var("HOME").unwrap().into();
            path.push(".btmsg");
            path
        };
    }

    /// Appends the given Blocktree path to the path of the given directory.
    fn socket_path(fs_path: &mut PathBuf, addr: &BlockAddr) {
        fs_path.push(addr.inode.to_string());
    }

    #[derive(Serialize, Deserialize, PartialEq, Eq, Default, Hash, Clone, Debug)]
    pub struct BlockAddr {
        pub root: Principal,
        pub generation: u64,
        pub inode: u64,
    }

    impl BlockAddr {
        pub fn new(root: Principal, generation: u64, inode: u64) -> Self {
            Self {
                root,
                generation,
                inode,
            }
        }
    }

    /// Generates and returns a new message ID.
    fn rand_msg_id() -> Result<u128> {
        const LEN: usize = std::mem::size_of::<u128>();
        let bytes = rand_array::<LEN>()?;
        let option = u128::read_from(bytes.as_slice());
        // Safety: because LEN == size_of::<u128>(), read_from should have returned Some.
        Ok(option.unwrap())
    }

    #[derive(Serialize, Deserialize, PartialEq, Eq, Hash, Clone)]
    pub struct Msg<T> {
        pub to: BlockAddr,
        pub from: BlockAddr,
        pub id: u128,
        pub body: T,
    }

    impl<T> Msg<T> {
        pub fn new(to: BlockAddr, from: BlockAddr, id: u128, body: T) -> Self {
            Self { to, from, id, body }
        }

        pub fn with_rand_id(to: BlockAddr, from: BlockAddr, body: T) -> Result<Self> {
            Ok(Self {
                to,
                from,
                id: rand_msg_id()?,
                body,
            })
        }
    }

    #[derive(Serialize, Deserialize)]
    enum VerMsg<T> {
        V0(Msg<T>),
    }

    /// A type which can be used to send messages.
    pub trait Sender {
        type SendFut<'a, T>: 'a + Future<Output = Result<()>> + core::marker::Send
        where
            Self: 'a,
            T: 'a + Serialize + core::marker::Send;

        fn send<'a, T: 'a + Serialize + core::marker::Send>(
            &'a mut self,
            msg: Msg<T>,
        ) -> Self::SendFut<'a, T>;

        fn addr(&self) -> &BlockAddr;

        fn send_msg<'a, T: 'a + Serialize + core::marker::Send>(
            &'a mut self,
            to: BlockAddr,
            body: T,
        ) -> Self::SendFut<'a, T> {
            let msg = Msg::with_rand_id(to, self.addr().clone(), body).unwrap();
            self.send(msg)
        }
    }

    /// A type which can be used to receive messages.
    pub trait Receiver<T>: Stream<Item = Result<Msg<T>>> {
        fn addr(&self) -> &BlockAddr;
    }

    /// Encodes and decodes messages using [btserde].
    struct MsgEncoder;

    impl<T: Serialize> Encoder<Msg<T>> for MsgEncoder {
        type Error = btlib::Error;

        fn encode(&mut self, item: Msg<T>, dst: &mut BytesMut) -> Result<()> {
            const U64_LEN: usize = std::mem::size_of::<u64>();
            let payload = dst.split_off(U64_LEN);
            let mut writer = payload.writer();
            write_to(&VerMsg::V0(item), &mut writer)?;
            let payload = writer.into_inner();
            let payload_len = payload.len() as u64;
            let mut writer = dst.writer();
            write_to(&payload_len, &mut writer)?;
            let dst = writer.into_inner();
            dst.unsplit(payload);
            Ok(())
        }
    }

    struct MsgDecoder<T>(PhantomData<T>);

    impl<T: for<'de> Deserialize<'de>> Decoder for MsgDecoder<T> {
        type Item = Msg<T>;
        type Error = btlib::Error;

        fn decode(&mut self, src: &mut BytesMut) -> Result<Option<Self::Item>> {
            let mut slice: &[u8] = src.as_ref();
            let payload_len: u64 = match read_from(&mut slice) {
                Ok(payload_len) => payload_len,
                Err(err) => {
                    if let btserde::Error::Eof = err {
                        return Ok(None);
                    }
                    return Err(err.into());
                }
            };
            let payload_len: usize = payload_len.try_into().box_err()?;
            if slice.len() < payload_len {
                src.reserve(payload_len - slice.len());
                return Ok(None);
            }
            let VerMsg::V0(msg) = read_from(&mut slice)?;
            // Consume all the bytes that have been read out of the buffer.
            let _ = src.split_to(std::mem::size_of::<u64>() + payload_len);
            Ok(Some(msg))
        }
    }

    /// Wraps a [UnixDatagram] and implements [AsyncRead] and [AsyncWrite] for it. Read operations
    /// are translated  to calls to `recv_from` and write operations are translated to `send`. Note
    /// that this means that writes will fail unless the wrapped socket is connected  to a peer.
    struct DatagramAdapter {
        socket: UnixDatagram,
    }

    impl DatagramAdapter {
        fn new(socket: UnixDatagram) -> Self {
            Self { socket }
        }

        fn get_ref(&self) -> &UnixDatagram {
            &self.socket
        }

        fn get_mut(&mut self) -> &mut UnixDatagram {
            &mut self.socket
        }
    }

    impl AsRef<UnixDatagram> for DatagramAdapter {
        fn as_ref(&self) -> &UnixDatagram {
            self.get_ref()
        }
    }

    impl AsMut<UnixDatagram> for DatagramAdapter {
        fn as_mut(&mut self) -> &mut UnixDatagram {
            self.get_mut()
        }
    }

    impl AsyncRead for DatagramAdapter {
        fn poll_read(
            self: Pin<&mut Self>,
            cx: &mut Context<'_>,
            buf: &mut ReadBuf<'_>,
        ) -> Poll<io::Result<()>> {
            self.socket.poll_recv(cx, buf)
        }
    }

    impl AsyncWrite for DatagramAdapter {
        fn poll_write(
            self: Pin<&mut Self>,
            cx: &mut Context<'_>,
            buf: &[u8],
        ) -> Poll<io::Result<usize>> {
            self.socket.poll_send(cx, buf)
        }

        fn poll_shutdown(self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<io::Result<()>> {
            Poll::Ready(self.socket.shutdown(Shutdown::Write))
        }

        fn poll_flush(self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<io::Result<()>> {
            Poll::Ready(Ok(()))
        }
    }

    /// An implementation of [Receiver] which uses a Unix datagram socket for receiving messages.
    struct UnixReceiver<T> {
        addr: BlockAddr,
        socket: FramedRead<DatagramAdapter, MsgDecoder<T>>,
    }

    impl<T: for<'de> Deserialize<'de>> UnixReceiver<T> {
        fn new(mut fs_path: PathBuf, addr: BlockAddr) -> Result<Self> {
            socket_path(&mut fs_path, &addr);
            let socket = DatagramAdapter::new(UnixDatagram::bind(fs_path)?);
            let socket = FramedRead::new(socket, MsgDecoder(PhantomData));
            Ok(Self { addr, socket })
        }
    }

    impl<T: for<'de> Deserialize<'de>> Stream for UnixReceiver<T> {
        type Item = Result<Msg<T>>;

        fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
            self.socket.poll_next_unpin(cx)
        }
    }

    impl<T: for<'de> Deserialize<'de>> Receiver<T> for UnixReceiver<T> {
        fn addr(&self) -> &BlockAddr {
            &self.addr
        }
    }

    /// Returns a [Receiver] which can be used to receive messages addressed to the given path.
    /// The `fs_path` argument specifies the filesystem directory under which the receiver's socket
    /// will be stored.
    pub fn local_receiver<T: for<'de> Deserialize<'de>>(
        fs_path: PathBuf,
        addr: BlockAddr,
    ) -> Result<impl Receiver<T>> {
        UnixReceiver::new(fs_path, addr)
    }

    /// An implementation of [Sender] which uses a Unix datagram socket to send messages.
    struct UnixSender {
        addr: BlockAddr,
        socket: FramedWrite<DatagramAdapter, MsgEncoder>,
    }

    impl UnixSender {
        fn new(mut fs_path: PathBuf, addr: BlockAddr) -> Result<Self> {
            let socket = UnixDatagram::unbound()?;
            socket_path(&mut fs_path, &addr);
            socket.connect(fs_path)?;
            let socket = FramedWrite::new(DatagramAdapter::new(socket), MsgEncoder);
            Ok(Self { addr, socket })
        }
    }

    impl<T: Serialize> Sink<Msg<T>> for UnixSender {
        type Error = btlib::Error;

        fn poll_ready(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<()>> {
            <tokio_util::codec::FramedWrite<DatagramAdapter, MsgEncoder> as futures::SinkExt<
                Msg<T>,
            >>::poll_ready_unpin(&mut self.socket, cx)
        }

        fn start_send(mut self: Pin<&mut Self>, item: Msg<T>) -> Result<()> {
            self.socket.start_send_unpin(item)
        }

        fn poll_flush(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<()>> {
            <tokio_util::codec::FramedWrite<DatagramAdapter, MsgEncoder> as futures::SinkExt<
                Msg<T>,
            >>::poll_flush_unpin(&mut self.socket, cx)
        }

        fn poll_close(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<()>> {
            <tokio_util::codec::FramedWrite<DatagramAdapter, MsgEncoder> as futures::SinkExt<
                Msg<T>,
            >>::poll_close_unpin(&mut self.socket, cx)
        }
    }

    impl Sender for UnixSender {
        fn addr(&self) -> &BlockAddr {
            &self.addr
        }

        type SendFut<'a, T>
            = Send<'a, FramedWrite<DatagramAdapter, MsgEncoder>, Msg<T>>
                where T: 'a + Serialize + core::marker::Send;

        fn send<'a, T: 'a + Serialize + core::marker::Send>(
            &'a mut self,
            msg: Msg<T>,
        ) -> Self::SendFut<'a, T> {
            self.socket.send(msg)
        }
    }

    /// Returns a [Sender] which can be used to send messages to the given Blocktree path.
    /// The `fs_path` argument specifies the filesystem directory in which to locate the
    /// socket of the recipient.
    pub fn local_sender(fs_path: PathBuf, addr: BlockAddr) -> Result<impl Sender> {
        UnixSender::new(fs_path, addr)
    }

    /// This is an identify function which allows you to specify a type parameter for the output
    /// of a future.
    /// This was needed to work around a failure in type inference for types with higher-rank
    /// lifetimes. Once this issue is resolved this can be removed:
    /// https://github.com/rust-lang/rust/issues/102211
    pub fn assert_send<'a, T>(
        fut: impl 'a + Future<Output = T> + core::marker::Send,
    ) -> impl 'a + Future<Output = T> + core::marker::Send {
        fut
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    use tempdir::TempDir;

    lazy_static! {
        static ref ROOT_PRINCIPAL: Principal =
            Principal::try_from("0!dSip4J0kurN5VhVo_aTipM-ywOOWrqJuRRVQ7aa-bew").unwrap();
    }

    fn block_addr(generation: u64, inode: u64) -> BlockAddr {
        BlockAddr::new(ROOT_PRINCIPAL.clone(), generation, inode)
    }

    #[derive(Serialize, Deserialize)]
    enum MsgError {
        Unknown,
    }

    #[derive(Deserialize)]
    enum BodyOwned {
        Ping,
        Success,
        Fail(MsgError),
        Read { offset: u64, size: u64 },
        Write { offset: u64, buf: Vec<u8> },
    }

    #[derive(Serialize)]
    enum BodyRef<'a> {
        Ping,
        Success,
        Fail(MsgError),
        Read { offset: u64, size: u64 },
        Write { offset: u64, buf: &'a [u8] },
    }

    struct TestCase {
        dir: TempDir,
    }

    impl TestCase {
        fn new() -> TestCase {
            Self {
                dir: TempDir::new("btmsg").unwrap(),
            }
        }

        fn endpoint(&self, inode: u64) -> (BlockAddr, impl Sender, impl Receiver<BodyOwned>) {
            let addr = block_addr(1, inode);
            let fs_path = self.dir.path().to_owned();
            let receiver = local_receiver(fs_path.clone(), addr.clone()).unwrap();
            let sender = local_sender(fs_path, addr.clone()).unwrap();
            (addr, sender, receiver)
        }
    }

    #[tokio::test]
    async fn message_received_is_message_sent() {
        let case = TestCase::new();
        let (addr, mut sender, mut receiver) = case.endpoint(1);

        sender.send_msg(addr.clone(), BodyRef::Ping).await.unwrap();
        let actual = receiver.next().await.unwrap().unwrap();

        let matched = if let BodyOwned::Ping = actual.body {
            true
        } else {
            false
        };
        assert!(matched);
        assert_eq!(&addr, &actual.to);
        assert_eq!(&addr, &actual.from);
    }

    #[tokio::test]
    async fn ping_pong() {
        let case = TestCase::new();
        let (addr_one, mut sender_one, mut receiver_one) = case.endpoint(1);
        let (addr_two, mut sender_two, mut receiver_two) = case.endpoint(2);

        let handle = tokio::spawn(async move {
            let msg = receiver_one.next().await.unwrap().unwrap();
            let reply_body = if let BodyOwned::Ping = msg.body {
                BodyRef::Success
            } else {
                BodyRef::Fail(MsgError::Unknown)
            };
            let fut = assert_send::<'_, Result<()>>(sender_two.send_msg(addr_one, reply_body));
            fut.await.unwrap();
        });

        sender_one.send_msg(addr_two, BodyRef::Ping).await.unwrap();
        handle.await.unwrap();
        let reply = receiver_two.next().await.unwrap().unwrap();
        let matched = if let BodyOwned::Success = reply.body {
            true
        } else {
            false
        };
        assert!(matched)
    }

    #[tokio::test]
    async fn read_write() {
        let case = TestCase::new();
        let (addr_one, mut sender_one, mut receiver_one) = case.endpoint(1);
        let (addr_two, mut sender_two, mut receiver_two) = case.endpoint(2);

        let handle = tokio::spawn(async move {
            let data: [u8; 8] = [0, 1, 2, 3, 4, 5, 6, 7];
            let msg = receiver_one.next().await.unwrap().unwrap();
            let reply_body = if let BodyOwned::Read { offset, size } = msg.body {
                let offset: usize = offset.try_into().unwrap();
                let size: usize = size.try_into().unwrap();
                let end: usize = offset + size;
                BodyRef::Write {
                    offset: offset as u64,
                    buf: &data[offset..end],
                }
            } else {
                BodyRef::Fail(MsgError::Unknown)
            };
            let msg = Msg::new(msg.from, addr_one, msg.id, reply_body);
            let fut = assert_send::<'_, Result<()>>(sender_two.send(msg));
            fut.await.unwrap();
        });

        sender_one
            .send_msg(addr_two, BodyRef::Read { offset: 2, size: 2 })
            .await
            .unwrap();
        handle.await.unwrap();
        let reply = receiver_two.next().await.unwrap().unwrap();
        if let BodyOwned::Write { offset, buf } = reply.body {
            assert_eq!(2, offset);
            assert_eq!([2, 3].as_slice(), buf.as_slice());
        } else {
            panic!("reply was not the right type");
        };
    }
}
